from enum import Enum


class TournamentState(Enum):
    """ State a tournament can be in """
    pending = 'pending'
    open_ = 'open'  #: can't use `open`
    complete = 'complete'
    in_progress = 'in progress'


class TournamentType(Enum):
    """ Type of a tournament """
    single_elimination = 'single elimination'
    double_elimination = 'double elimination'
    round_robin = 'round robin'
    swiss = 'swiss'


class TournamentStateResult(Enum):
    """ State given from the Challonge API.
    Can be different from :class:`TournamentState`
    """
    underway = 0
    pending = 1


class DoubleEliminationEnding(Enum):
    """ Type of ending for double elimination tournaments """
    default = None  #: give the winners bracket finalist two chances to beat the losers bracket finalist
    single_match = 'single_match'  #: create only one grand finals match
    no_grand_finals = 'skip'  #: don't create a finals match between winners and losers bracket finalists


class RankingOrder(Enum):
    """ Order the ranking should be built upon """
    match_wins = 'match wins'
    game_wins = 'game wins'
    points_scored = 'points scored'
    points_difference = 'points difference'
    custom = 'custom'


class Pairing(Enum):
    """ Method of participant pairing when building matches """
    seeds = 0
    sequential = 1


class MatchState(Enum):
    """ State a match can be in """
    all_ = 'all'  #: can't use `all`
    open_ = 'open'  #: can't use `open`
    pending = 'pending'
    complete = 'complete'



from . import AUTO_GET_PARTICIPANTS, AUTO_GET_MATCHES
from .helpers import get_connection
from .tournament import Tournament, TournamentType


class User:
    """ Representation of a Challonge user

    Main entry point for using the async challonge library.

    """

    def __init__(self, username: str, api_key: str, **kwargs):
        self.tournaments = None
        self.connection = get_connection(username, api_key, **kwargs)

    def _refresh_tournament_from_json(self, tournament_data):
        if self.tournaments is None:
            self.tournaments = [self._create_tournament(tournament_data)]
        else:
            for t in self.tournaments:
                if tournament_data['tournament']['id'] == t._id:
                    t._refresh_from_json(tournament_data)
                    break
            else:
                self.tournaments.append(self._create_tournament(tournament_data))

    def _create_tournament(self, json_def) -> Tournament:
        return Tournament(self.connection, json_def)

    def _find_tournament(self, e_id):
        if self.tournaments is not None:
            for e in self.tournaments:
                if e.id == int(e_id):
                    return e
        return None

    async def validate(self):
        """ checks whether the current user is connected

        |methcoro|

        Raises:
            APIException

        """
        await self.connection('GET', 'tournaments')

    async def get_tournament(self, t_id: int, force_update=False) -> Tournament:
        """ gets a tournament with its id

        |methcoro|

        Args:
            t_id: tournament id
            force_update: *optional* set to True to force the data update from Challonge

        Returns:
            Tournament

        Raises:
            APIException

        """
        found_t = self._find_tournament(t_id)
        if force_update or found_t is None:
            res = await self.connection('GET', 'tournaments/{}'.format(t_id))
            self._refresh_tournament_from_json(res)
            found_t = self._find_tournament(t_id)

        return found_t

    async def get_tournaments(self, force_update=False) -> list:
        """ gets all user's tournaments

        |methcoro|

        Args:
            force_update: *optional* set to True to force the data update from Challonge

        Returns:
            list[Tournament]: list of all the user tournaments

        Raises:
            APIException

        """
        if force_update or self.tournaments is None:
            params = {
                'include_participants': 1 if AUTO_GET_PARTICIPANTS else 0,
                'include_matches': 1 if AUTO_GET_MATCHES else 0
            }
            res = await self.connection('GET', 'tournaments', **params)
            if len(res) == 0:
                self.tournaments = []
            else:
                for t_data in res:
                    self._refresh_tournament_from_json(t_data)

        return self.tournaments

    async def create_tournament(self, name: str, url: str, tournament_type=TournamentType.single_elimination, **params) -> Tournament:
        """ creates a simple tournament with basic options

        |methcoro|

        Args:
            name: name of the new tournament
            url: url of the new tournament (http://challonge.com/url)
            tournament_type: Defaults to TournamentType.single_elimination
            params: optional params (see http://api.challonge.com/v1/documents/tournaments/create)

        Returns:
            Tournament: the newly created tournament

        Raises:
            APIException

        """
        params.update({
            'name': name,
            'url': url,
            'tournament_type': tournament_type.value,
        })
        res = await self.connection('POST', 'tournaments', 'tournament', **params)
        self._refresh_tournament_from_json(res)
        return self._find_tournament(res['tournament']['id'])

    async def destroy_tournament(self, t: Tournament):
        """ completely removes a tournament from Challonge

        |methcoro|

        Note:
            |from_api| Deletes a tournament along with all its associated records. There is no undo, so use with care!

        Raises:
            APIException

        """
        await self.connection('DELETE', 'tournaments/{}'.format(t.id))
        if t in self.tournaments:
            self.tournaments.remove(t)


async def get_user(username: str, api_key: str, **kwargs) -> User:
    """ Creates a new user, validate its credentials and returns it

    |funccoro|

    Args:
        username: username as specified on the challonge website
        api_key: key as found on the challonge
            `settings <https://challonge.com/settings/developer>`_

    Returns:
        User: a logged in user if no exception has been raised

    Raises:
        APIException

    """
    new_user = User(username, api_key, **kwargs)
    await new_user.validate()
    return new_user



import aiohttp
import logging

import challonge


DEFAULT_TIMEOUT = 30
log = logging.getLogger('challonge')


class APIException(Exception):
    """ If anything goes wrong during a request to the Challonge API, this exception will be raised. """
    pass


def assert_or_raise(cond, exc, *args):
    if challonge.USE_EXCEPTIONS is not None and not cond:
        if challonge.USE_EXCEPTIONS:
            raise exc(*args)
        else:
            log.warning('An exception `{}` has been raised: `{}`'.format(exc.__name__, args))


class FieldDescriptor:
    def __init__(self, attr):
        self.attr = attr
        self.__doc__ = 'readonly attribute'

    def __get__(self, instance, owner):
        return getattr(instance, self.attr) if instance else self


class FieldHolder(type):
    private_name = '_{}'

    def _create_holder(self, holder_class, json_def, **kwargs):
        return holder_class(self.connection, json_def, **kwargs)

    def _find_holder(self, local_list, e_id):
        if local_list is not None:
            for e in local_list:
                if e.id == int(e_id):
                    return e
        return None

    def _get_from_dict(self, data):
        for a in self._fields:
            name = FieldHolder.private_name.format(a) if challonge.USE_FIELDS_DESCRIPTORS else a
            setattr(self, name, data[a] if a in data else None)

    def __init__(cls, name, bases, dct):
        super(FieldHolder, cls).__init__(name, bases, dct)

        cls._create_holder = FieldHolder._create_holder
        cls._find_holder = FieldHolder._find_holder
        cls._get_from_dict = FieldHolder._get_from_dict
        cls.__eq__ = lambda self, other: self._id == other._id

        if challonge.USE_FIELDS_DESCRIPTORS:
            for a in cls._fields:
                setattr(cls, a, FieldDescriptor(FieldHolder.private_name.format(a)))


class Connection:
    challonge_api_url = 'https://api.challonge.com/v1/{}.json'

    def __init__(self, username: str, api_key: str, timeout, loop):
        self.username = username
        self.api_key = api_key
        self.timeout = timeout
        self.loop = loop

    async def __call__(self, method: str, uri: str, params_prefix: str =None, **params):
        """ response codes:
        200 - OK
        401 - Unauthorized (Invalid API key or insufficient permissions)
        404 - Object not found within your account scope
        406 - Requested format is not supported - request JSON or XML only
        422 - Validation error(s) for create or update method
        500 - Something went wrong on our end. If you continually receive this, please contact us.
        """
        params = self._prepare_params(params, params_prefix)
        # print(params)

        # build the HTTP request and use basic authentication
        url = self.challonge_api_url.format(uri)

        async with aiohttp.ClientSession(loop=self.loop) as session:
            with aiohttp.Timeout(self.timeout):
                auth = aiohttp.BasicAuth(login=self.username, password=self.api_key)
                async with session.request(method, url, params=params, auth=auth) as response:
                    resp = await response.json()
                    assert_or_raise(response.status in [200, 401, 404, 406, 422, 500], ValueError, 'Unknown API return code', response.status, response.reason, uri, params)
                    assert_or_raise(response.status not in [401, 404, 406, 422, 500], APIException, response.status, response.reason, uri, params)
                    return resp

    @staticmethod
    def _prepare_params(params, prefix=None) -> dict:
        def val(value):
            if isinstance(value, bool):
                # challonge only accepts lowercase true/false
                value = str(value).lower()
            return str(value)

        new_params = []
        if prefix:
            if prefix.endswith('[]'):
                for k, v in params.items():
                    new_params.extend([('{}[{}]'.format(prefix, k), val(u)) for u in v])
            else:
                new_params = [('{}[{}]'.format(prefix, k), val(v)) for k, v in params.items()]
        else:
            new_params = [(k, val(v)) for k, v in params.items()]
        return new_params


def get_connection(username, api_key, timeout=DEFAULT_TIMEOUT, loop=None):
    return Connection(username, api_key, timeout, loop)



from .helpers import FieldHolder
from .enums import MatchState


class Participant(metaclass=FieldHolder):
    """ Representation of a Challonge participant """

    _fields = ['active', 'checked_in_at', 'created_at', 'final_rank',
               'group_id', 'icon', 'id', 'invitation_id', 'invite_email',
               'misc', 'name', 'on_waiting_list', 'seed', 'tournament_id',
               'updated_at', 'challonge_username', 'challonge_email_address_verified',
               'removable', 'participatable_or_invitation_attached', 'confirm_remove',
               'invitation_pending', 'display_name_with_invitation_email_address',
               'email_hash', 'username', 'attached_participatable_portrait_url',
               'can_check_in', 'checked_in', 'reactivatable',
               'display_name', 'group_player_ids']

    def __init__(self, connection, json_def, tournament, **kwargs):
        self.connection = connection
        self._tournament = tournament
        self._refresh_from_json(json_def)

    def _refresh_from_json(self, json_def):
        if 'participant' in json_def:
            self._get_from_dict(json_def['participant'])

    async def _change(self, **params):
        res = await self.connection('PUT',
                                    'tournaments/{}/participants/{}'.format(self._tournament_id, self._id),
                                    'participant',
                                    **params)
        self._refresh_from_json(res)

    async def change_display_name(self, new_name: str):
        """ Change the name displayed on the Challonge website

        |methcoro|

        Note:
            |from_api| The name displayed in the bracket/schedule. Must be unique per tournament.

        Args:
            new_name: as described

        Raises:
            APIException

        """
        await self._change(name=new_name)

    async def change_username(self, username: str):
        """ Will invite the Challonge user to the tournament

        |methcoro|

        Args:
            username: Challonge username

        Raises:
            APIException

        """
        await self._change(challonge_username=username)

    async def change_email(self, email: str):
        """  Set / update the email associated to the participant

        |methcoro|

        Setting an email will first search for a matching Challonge account.
        If one is found, it will be invited to the tournament.
        If one is not found, the "new-user-email" attribute will be set, and the user will be invited via email to create an account.

        Args:
            email: as described

        Raises:
            APIException

        """
        await self._change(email=email)

    async def change_seed(self, new_seed: int) -> int:
        """ Change the seed of the participant

        |methcoro|

        Note:
            |from_api| Must be between 1 and the current number of participants (including the new record). Overwriting an existing seed will automatically bump other participants as you would expect.

        Args:
            new_seed: as described

        Returns:
            the same seed number as passed or `None` if something failed

        Raises:
            APIException

        """
        await self._change(seed=new_seed)

    async def change_misc(self, misc: str) -> str:
        """ Change the `misc` field

        |methcoro|

        Note:
            |from_api| Max: 255 characters. Multi-purpose field that is only visible via the API and handy for site integration (e.g. key to your users table)

        Args:
            misc: str content

        Raises:
            APIException

        """
        await self._change(misc=misc)

    async def check_in(self):
        """ Checks this participant in

        |methcoro|

        Warning:
            |unstable|

        Raises:
            APIException

        """
        res = await self.connection('POST', 'tournaments/{}/participants/{}/check_in'.format(self._tournament_id, self._id))
        self._refresh_from_json(res)

    async def undo_check_in(self):
        """ Undo the check in for this participant

        |methcoro|

        Warning:
            |unstable|

        Raises:
            APIException

        """
        res = await self.connection('POST', 'tournaments/{}/participants/{}/undo_check_in'.format(self._tournament_id, self._id))
        self._refresh_from_json(res)

    async def get_matches(self, state: MatchState = MatchState.all_):
        """ Return the matches of the given state

        |methcoro|

        Args:
            state: see :class:`MatchState`

        Raises:
            APIException

        """
        matches = await self.connection('GET',
                                        'tournaments/{}/matches'.format(self._tournament_id),
                                        state=state.value,
                                        participant_id=self._id)
        # return [await self._tournament.get_match(m['match']['id']) for m in matches] 3.6 only...
        ms = []
        for m in matches:
            ms.append(await self._tournament.get_match(m['match']['id']))
        return ms

    async def get_next_match(self):
        """ Return the first open match found, or if none, the first pending match found

        |methcoro|

        Raises:
            APIException

        """
        if self._final_rank is not None:
            return None

        matches = await self.get_matches(MatchState.open_)

        if len(matches) == 0:
            matches = await self.get_matches(MatchState.pending)

        if len(matches) > 0:
            return matches[0]

        return None

    async def get_next_opponent(self):
        """ Get the opponent of the potential next match. See :func:`get_next_match`

        |methcoro|

        Raises:
            APIException

        """
        next_match = await self.get_next_match()
        if next_match is not None:
            opponent_id = next_match.player1_id if next_match.player2_id == self._id else next_match.player2_id
            return await self._tournament.get_participant(opponent_id)
        return None
